/* -*- mode: C++; c-basic-offset: 2; -*- */
#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

#include <iostream>
#include <set>
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>

#include "general/AtExit.hh"
#include "general/autoarray.hh"
#include "general/fstream.hh"
#include "general/gpstime.hh"
#include "general/types.hh"

#include "framecpp/Common/FrameSpec.hh"
#include "framecpp/Common/CheckSum.hh"
#include "framecpp/Common/IOStream.hh"
#include "framecpp/Common/FrameStream.hh"

#include "framecpp/FrameCPP.hh"

#include "framecpp/FrameH.hh"
#include "framecpp/FrAdcData.hh"
#include "framecpp/FrRawData.hh"
#include "framecpp/FrVect.hh"

#include "framecpp/Dimension.hh"

#include "FrSample.hh"
#include "FrSample3.tcc"
#include "FrSample4.tcc"
#include "FrSample6.tcc"
#include "FrSample7.tcc"
#include "FrSample8.tcc"

using FrameCPP::Common::CheckSum;
using FrameCPP::Common::FrameBuffer;
using FrameCPP::Common::OFrameStream;
using FrameCPP::Common::OStream;
using General::filebuf;

#include "test.hh"

using std::string;
using std::ostringstream;
using General::GPSTime;

using FrameCPP::Dimension;

using FrameCPP::FrameH;
using FrameCPP::FrAdcData;
using FrameCPP::FrRawData;
using FrameCPP::FrVect;

// The name of the program executable
static char* programname = 0;

typedef std::set<std::string> channel_set_type;

//=======================================================================
// Forward declaration of local functions
//=======================================================================

//-----------------------------------------------------------------------
static void
all_type_frame( const string& Type,
		const channel_set_type& ChannelSet,
		const GPSTime& Start,
		const INT_4U DeltaT,
		const INT_4U Iterations );

//-----------------------------------------------------------------------

static void
all_type_iteration( const INT_4U DeltaT,
		    const channel_set_type& ListOfChannels,
		    std::auto_ptr< FrameH >& Frame,
		    INT_4U* Channel,
		    INT_4U Iteration );

//-----------------------------------------------------------------------

template< class T >
static std::string
data_type( )
{
  return "";
}

//-----------------------------------------------------------------------

template< class T >
static std::auto_ptr< FrAdcData >
ramped_channel( INT_4U DeltaT, INT_4U* Channel, INT_4U Iteration,
		T Start, T Stop, T Inc,
		INT_4U SampleRate );

//-----------------------------------------------------------------------

template< class T >
static T* ramp_data_create( INT_4U Samples, T Start, T Stop, T Inc );

//-----------------------------------------------------------------------

template< class T >
static string ramped_name( INT_4U Iteration );

//-----------------------------------------------------------------------

template< class T >
static std::auto_ptr< FrVect >
ramped_vector( INT_4U DeltaT, INT_4U SampleRate, INT_4U Iteration,
	       T Start, T Stop, T Inc );

//-----------------------------------------------------------------------

static void usage( char* ProgramName, int Exit );

//-----------------------------------------------------------------------

void
test_frame( int Version )
{
  using FrameCPP::Common::FrameSpec;

  //-------------------------------------------------------------------
  // Create the frame
  //-------------------------------------------------------------------
  std::list< FrameSpec::Object* >	stat_data;
  std::auto_ptr< FrameSpec::Object >	frame;

  switch( Version )
  {
  case 3:
    frame = makeFrame< 3 >( stat_data );
    break;
  case 4:
    frame = makeFrame< 4 >( stat_data );
    break;
  case 6:
    frame = makeFrame< 6 >( stat_data );
    break;
  case 7:
    frame = makeFrame< 7 >( stat_data );
    break;
  case 8:
    frame = makeFrame< 8 >( stat_data );
    break;
  default:
    {
      std::ostringstream msg;

      msg << "name makeFrame case for version: " << Version;
      throw std::runtime_error( msg.str( ) );
    }
    break;
  }
  //-------------------------------------------------------------------
  // Calculate the filename
  //-------------------------------------------------------------------
  std::ostringstream	filename;

  filename << "Z-R_std_test_frame_ver" << Version
	   << "-" << 600000000
	   << "-" << 1
	   << ".gwf"
    ;

  if ( frame.get( ) == (FrameSpec::Object*)NULL )
  {
    throw std::runtime_error( "Failed to create frame" );
  }
  //-------------------------------------------------------------------
  // Creation of io buffer.
  //-------------------------------------------------------------------
  FrameBuffer< filebuf >*
    obuf( new FrameBuffer< filebuf >( std::ios::out ) );
  try
  {
    //-----------------------------------------------------------------
    // Open the buffer and the stream
    //-----------------------------------------------------------------
    obuf->open( filename.str( ).c_str( ),
		std::ios::out | std::ios::binary );

    OFrameStream	ofs( obuf, Version );
  
    ofs.SetCheckSumFile( CheckSum::CRC );

    //-----------------------------------------------------------------
    // Write the FrStatData data to the stream
    //-----------------------------------------------------------------
    for ( std::list< FrameSpec::Object* >::iterator
	    cur = stat_data.begin( ),
	    last = stat_data.end( );
	  cur != last;
	  ++cur )
    {
      ofs.WriteFrStatData( *(*cur) );
    }
    //-----------------------------------------------------------------
    // Write the frame to the file
    //-----------------------------------------------------------------
    ofs.WriteFrame( *(frame), CheckSum::CRC );
    //-----------------------------------------------------------------
    // Close the stream and file
    //-----------------------------------------------------------------
    ofs.Close( );
    obuf->close( );
  }
  catch( ... )
  {
    obuf->close( );
    // unlink( filename.str( ).c_str( ) );
    throw;
  }

  //-------------------------------------------------------------------
  // Release resources associated with the FrStatData
  //-------------------------------------------------------------------
  for ( std::list< FrameSpec::Object* >::iterator
	  cur = stat_data.begin( ),
	  last = stat_data.end( );
	cur != last;
	++cur )
  {
    delete *cur;
  }
  stat_data.clear( );
}

int
main(int argc, char* argv[])
  try
  {
    bool	create_test_frame = true;
    int		args_parsed;

    INT_4U		deltat = 1;
    INT_4U		iterations = 1;
    GPSTime		start( 600000000, 0 );
    string		type( "ilwd_test_frame" );
    channel_set_type	channel_set;
    int			version = FRAME_SPEC_CURRENT;

    /* Determine program name (strip filesystem path) */
    for (programname = argv [0] + strlen (argv [0]) - 1;
	 programname != argv [0] && *programname != '/';
	 programname--)
      ;
    
    if (*programname == '/')
    {
      programname++;
    }

    for ( args_parsed = 1; args_parsed < argc; ++args_parsed )
    {
      char* cur_arg = argv[ args_parsed ];
	
      if ( strlen( cur_arg ) > 2 )
      {
	if ( strncmp( cur_arg, "-c", 2 ) == 0 )
	{
	  std::string	opt( &cur_arg[2] );

	  // Convert to lower case
	  for( std::string::iterator
		 cur = opt.begin( ),
		 last = opt.end( );
	       cur != last;
	       ++cur )
	  {
	    *cur = tolower( *cur );
	  }
	  channel_set.insert( opt );
	  create_test_frame = false;
	}
	else if ( strncmp( cur_arg, "-d", 2 ) == 0 )
	{
	  // Specify the duration in seconds for the output data.
	  deltat = atol( &(cur_arg[2]) );
	  create_test_frame = false;
	}
	else if ( ! strncasecmp( cur_arg, "-h", 2 ) 
		  || ! strncasecmp( cur_arg, "--h", 3 ) )
	{
	  usage( argv[0], 0 );
	}
	else if ( strncmp( cur_arg, "-i", 2 ) == 0 )
	{
	  // Specify the number of iterations
	  iterations = atol( &(cur_arg[2]) );
	  create_test_frame = false;
	}
	else if ( strncmp( cur_arg, "-s", 2 ) == 0 )
	{
	  // Specify the gps start time in whole seconds
	  start = GPSTime( atol( &(cur_arg[2]) ), 0 );
	  create_test_frame = false;
	}
	else if ( strncmp( cur_arg, "-t", 2 ) == 0 )
	{
	  // Specify the gps start time in whole seconds
	  type = &(cur_arg[2]);
	  create_test_frame = false;
	}
	else if ( strncmp( cur_arg,
			   "--version=", sizeof( "--version=" ) - 1 )
		  == 0 )
	{
	  version = atol( &cur_arg[ sizeof( "--version=" ) - 1 ] );
	}
	else
	{
	  break;
	}
      }
      else
      {
	break;
      }
    }
    //  usage help
    if (args_parsed != argc )
    {
      usage( argv[0], 1 );
    }

    if ( create_test_frame )
    {
      test_frame( version );
    }
    else
    {
      if ( channel_set.size( ) == 0 )
      {
	channel_set.insert( "int_2u" );
	channel_set.insert( "int_2s" );
	channel_set.insert( "int_4u" );
	channel_set.insert( "int_4s" );
	channel_set.insert( "int_8u" );
	channel_set.insert( "int_8s" );
	channel_set.insert( "real_4" );
	channel_set.insert( "real_8" );
	channel_set.insert( "complex_8" );
	channel_set.insert( "complex_16" );
      }
      all_type_frame( type, channel_set, start, deltat, iterations );
    }

    General::AtExit::Cleanup( );
    return 0;
  } catch ( const std::exception& e ) {
    std::cerr << "FATAL: Caught exception: " << e.what( ) << std::endl;
    General::AtExit::Cleanup( );
    return 1;
  } catch (...) {
    std::cerr << "FATAL: Caught unknown exception: " << std::endl;
    General::AtExit::Cleanup( );
    return 1;
  }

//=======================================================================
// Local functions
//=======================================================================

//-----------------------------------------------------------------------
//
//-----------------------------------------------------------------------
static void
all_type_frame( const string& Type,
		const channel_set_type& ChannelSet,
		const GPSTime& Start,
		const INT_4U DeltaT,
		const INT_4U Iterations )
{
  //-------------------------------------------------------------------
  // Create filename
  //-------------------------------------------------------------------
  std::ostringstream	filename;

  filename << "Z-" << Type << "-"
	   << Start.GetSeconds( ) << "-" << DeltaT
	   << ".gwf";

  //-------------------------------------------------------------------
  // Open the buffer and the stream
  //-------------------------------------------------------------------
  FrameBuffer< filebuf >*
    obuf( new FrameBuffer< filebuf >( std::ios::out ) );
  obuf->open( filename.str( ).c_str( ),
	      std::ios::out | std::ios::binary );

  OFrameStream	ofs( obuf );
  
  ofs.SetCheckSumFile( CheckSum::CRC );

  //-------------------------------------------------------------------
  // Generation of the frame
  //-------------------------------------------------------------------

  std::auto_ptr< FrameH >
    frame( new FrameH( Type, -1, 0,
		       Start, Start.GetLeapSeconds( ),
		       REAL_8( DeltaT ) ) );

  frame->SetRawData( new FrRawData( "RawData" ),
		     false, /* allocate */
		     true /* own */ );

  //-------------------------------------------------------------------
  // Loop over the number of iterations for creation of the data sets
  //-------------------------------------------------------------------

  for ( INT_4U i = 1; i <= Iterations; ++i )
  {
    INT_4U	channel = 0;
    all_type_iteration( DeltaT, ChannelSet, frame, &channel, i );
  }

  //-------------------------------------------------------------------
  // Writing of the frame
  //-------------------------------------------------------------------

  ofs.WriteFrame( *(frame.get( )) );

  //-------------------------------------------------------------------
  // Close the frame and file stream
  //-------------------------------------------------------------------
  ofs.Close();
  obuf->close( );
}

//-----------------------------------------------------------------------
//
//-----------------------------------------------------------------------

static void
all_type_iteration( const INT_4U DeltaT,
		    const channel_set_type& ListOfChannels,
		    std::auto_ptr< FrameH >& Frame,
		    INT_4U* Channel,
		    INT_4U Iteration )
{
  FrRawData*	rd( Frame->GetRawData( ) );

  //-------------------------------------------------------------------
  // Make sure that we do not dereference a null pointer
  //-------------------------------------------------------------------
  if ( rd == (FrRawData*)NULL )
  {
    return;
  }
  //-------------------------------------------------------------------
  // Safe to use
  //-------------------------------------------------------------------
  FrRawData::firstAdc_type& adclist( rd->RefFirstAdc( ) );
  //-------------------------------------------------------------------
  // Add the ramped channels
  //-------------------------------------------------------------------
  channel_set_type::const_iterator	last_channel( ListOfChannels.end( ) );

  if ( ListOfChannels.find( "int_2u" ) != last_channel )
  {
    adclist.append( ramped_channel< INT_2U >( DeltaT, Channel, Iteration,
					      0, 512, 1,
					      512 ).release( ),
		    false /* allocate */,
		    true /* own */ );
  }
  if ( ListOfChannels.find( "int_2s" ) != last_channel )
  {
    adclist.append( ramped_channel< INT_2S >( DeltaT, Channel, Iteration,
					      -512, 512, 1,
					      512 ).release( ),
		    false /* allocate */,
		    true /* own */ );
  }
  if ( ListOfChannels.find( "int_4u" ) != last_channel )
  {
    adclist.append( ramped_channel< INT_4U >( DeltaT, Channel, Iteration,
					      0, 1024, 1,
					      1024 ).release( ),
		    false /* allocate */,
		    true /* own */ );
  }
  if ( ListOfChannels.find( "int_4s" ) != last_channel )
  {
    adclist.append( ramped_channel< INT_4S >( DeltaT, Channel, Iteration,
					      -1024, 1024, 1,
					      1024 ).release( ),
		    false /* allocate */,
		    true /* own */ );
  }
  if ( ListOfChannels.find( "int_8u" ) != last_channel )
  {
    adclist.append( ramped_channel< INT_8U >( DeltaT, Channel, Iteration,
					      0, 2048, 1,
					      1024 ).release( ),
		    false /* allocate */,
		    true /* own */ );
  }
  if ( ListOfChannels.find( "int_8s" ) != last_channel )
  {
    adclist.append( ramped_channel< INT_8S >( DeltaT, Channel, Iteration,
					      -2048, 2048, 1,
					      1024 ).release( ),
		    false /* allocate */,
		    true /* own */ );
  }

  if ( ListOfChannels.find( "real_4" ) != last_channel )
  {
    adclist.append( ramped_channel< REAL_4 >( DeltaT, Channel, Iteration,
					      -1024, 1024, 0.5,
					      1024 ).release( ),
		    false /* allocate */,
		    true /* own */ );
  }
  if ( ListOfChannels.find( "real_8" ) != last_channel )
  {
    adclist.append( ramped_channel< REAL_8 >( DeltaT, Channel, Iteration,
					      -1024, 1024, 0.5,
					      1024 ).release( ),
		    false /* allocate */,
		    true /* own */ );
  }

  if ( ListOfChannels.find( "complex_8" ) != last_channel )
  {
    adclist.append( ramped_channel< COMPLEX_8 >( DeltaT, Channel, Iteration,
						 COMPLEX_8( -1024, 0),
						 COMPLEX_8( 1024, 0),
						 COMPLEX_8( 1, 0 ),
						 1024 ).release( ),
		    false /* allocate */,
		    true /* own */ );
  }
  if ( ListOfChannels.find( "complex_16" ) != last_channel )
  {
    adclist.append( ramped_channel< COMPLEX_16 >( DeltaT, Channel, Iteration,
						  COMPLEX_16( -1024, 0),
						  COMPLEX_16( 1024, 0),
						  COMPLEX_16( 1, 0 ),
						  1024 ).release( ),
		    false /* allocate */,
		    true /* own */ );
  }

}

#define DATA_TYPE(A) template< >		\
  std::string					\
  data_type< A >( )				\
  {						\
    return #A;					\
  }

DATA_TYPE(CHAR_U)
DATA_TYPE(CHAR)
DATA_TYPE(INT_2U)
DATA_TYPE(INT_2S)
DATA_TYPE(INT_4U)
DATA_TYPE(INT_4S)
DATA_TYPE(INT_8U)
DATA_TYPE(INT_8S)
DATA_TYPE(REAL_4)
DATA_TYPE(REAL_8)
DATA_TYPE(COMPLEX_8)
DATA_TYPE(COMPLEX_16)

#undef DATA_TYPE

//-----------------------------------------------------------------------
//
//-----------------------------------------------------------------------

template< class T >
static std::auto_ptr< FrAdcData >
ramped_channel( INT_4U DeltaT, INT_4U* Channel, INT_4U Iteration,
		T Start, T Stop, T Inc,
		INT_4U SampleRate )
{
  //-------------------------------------------------------------------
  // Create the FrAdcData structure to hold the Adc metadata
  //-------------------------------------------------------------------
  std::auto_ptr< FrAdcData >
    adc ( new FrAdcData( ramped_name< T >( Iteration ),
				   0, /* Group */
				   *Channel, /* Channel Number */
				   32, /* NBits */
				   SampleRate, /* sampleRate */
				   0.0, /* bias */
				   1.0, /* slope */
				   "counts", /* units */
				   0.0, /* fShift */
				   0.0, /* timeOffset */
				   0, /* dataValid */
				   0.0 /* phase */ ) );
  //-------------------------------------------------------------------
  // Link the data to the metadata
  //-------------------------------------------------------------------
  adc->RefData( ).append( ramped_vector< T >( DeltaT,
					      SampleRate,
					      Iteration,
					      Start, Stop, Inc
					      ).release( ),
			  false, /* allocate */
			  true /* own */ );
  //-------------------------------------------------------------------
  // Incriment the channel number
  //-------------------------------------------------------------------
  (*Channel)++;
  return adc;
}

//-----------------------------------------------------------------------
//
//-----------------------------------------------------------------------

template< class T >
static T*
ramp_data_create( INT_4U Samples, T Start, T Stop, T Inc )
{
  General::AutoArray< T > data( new T[ Samples ] );
  data[ 0 ] = Start;
  for ( INT_4U x = 1; x < Samples; x++ )
  {
    if ( data[x -1] == Stop )
    {
      data[x] = Start;
    }
    else
    {
      data[ x ] = data[ x - 1 ] + Inc;
    }
  }
  return data.release( );
}

//-----------------------------------------------------------------------
//
//-----------------------------------------------------------------------

template< class T >
static string
ramped_name( INT_4U Iteration )
{
  ostringstream	name;

  name << "Z0:RAMPED_" << data_type< T >( ) << "_" << Iteration;
  return name.str( );
}

//-----------------------------------------------------------------------
//
//-----------------------------------------------------------------------

template< class T >
static std::auto_ptr< FrVect >
ramped_vector( INT_4U DeltaT, INT_4U SampleRate, INT_4U Iteration,
	       T Start, T Stop, T Inc )
{
  const INT_4U samples( DeltaT * SampleRate );

  Dimension	dim( samples );
  General::AutoArray< T > data( ramp_data_create< T >( samples,
						       Start,
						       Stop,
						       Inc ) );

  std::auto_ptr< FrVect >
    vect( new FrVect( ramped_name< T >( Iteration ), 1, &dim,
				data.release( ), "" /* unitY */,
				false /* allocate */, true /* own */ ) );
  return vect;
}

//-----------------------------------------------------------------------
//
//-----------------------------------------------------------------------

static void
usage( char* ProgramName, int Exit )
{
  std::cout << "usage: " << ProgramName
	    << " [-c<channel type>]"
	    << " [-d<duration>]"
	    << " [-h]"
	    << " [-i<iterations>]"
	    << " [-s<start time>]"
	    << " [-t<type specifier>]"
	    << std::endl
#if 0
	    << "    With no options specified, create LIGO standard testing"
	    << std::endl
	    << "    frame file '" << stdTestFrameName << "'"
#endif /* 0 */
	    << std::endl
	    << std::endl
	    << "    If any options are specified, then a frame file with"
	    << std::endl
	    << "    one or more datatypes is generated where:"
	    << std::endl
	    << "        <channel type> is the type of channel" << std::endl
	    << "        <duration> is frame length in seconds" << std::endl
	    << "        <iterations> is the number of sets of channels to produce" << std::endl
	    << "        <start time> specifies the gps start time for the frame" << std::endl
	    << "        <type specifier> is the type component of the frame file name" << std::endl;
  General::AtExit::Cleanup( );
  exit( Exit );
}

